//*****************************************************************************
//* =====================
//*  AZSpring
//* =====================
//*  
//* (Description)
//*    e+e- --> AZ generator
//*
//* (Update Record)
//*    2017/10/14  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "AZSpring.h"

#include <sstream>
#include <iomanip>
//#define __DEBUG__
//#define __ZEROWIDTH__
//#define __PHASESPACE__
//#define __NODECAY__
#ifdef __PHASESPACE__
#define __NODECAY__
#endif

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(AZSpring)
ClassImp(AZSpringBuf)
ClassImp(AZBases)

//-----------------------------------------------------------------------------
// ==============================
//  class AZSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
AZSpring::AZSpring(const char    *name,
                   const char    *title,
                         AZBases *bases)
         : JSFSpring(name, title, bases)
{
  fEventBuf = new AZSpringBuf("AZSpringBuf",
                              "AZSpring event buffer",
                              this);
  if (!bases) { 
    SetBases(new AZBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
AZSpring::~AZSpring()
{
  delete fEventBuf;
  delete GetBases();
}


//-----------------------------------------------------------------------------
// ==============================
//  class AZSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t AZSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  AZBases    *bases   = static_cast<AZBases*>(
		        static_cast<AZSpring*>(Module())->GetBases());

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  const Int_t kNparton = 4;
  ANL4DVector pv[kNparton];
  pv[0] = bases->fP[0];  // photon
  pv[2] = bases->fP[1];  // fz1
  pv[3] = bases->fP[2];  // fz2
  pv[1] = pv[2] + pv[3]; // Z

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fQ2AZ         = fEcmIP*fEcmIP;
  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();
  fQ2Z          = bases->GetQ2Z();
  fCosThetaF    = bases->GetCosThetaF();
  fPhiF         = bases->GetPhiF();
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);
  for (Int_t i=0; i<kNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  fNparton = kNparton;

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [kNparton];
    for (Int_t i=0; i<kNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<kNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }

  Int_t    ida     = 22;                          // PDG code for photon
  Int_t    idz     = 23;                          // PDG code for Z
  Int_t    idf     = bases->f1Ptr->GetPID   ();   // PDG code for f
  Double_t chrg    = bases->f1Ptr->GetCharge();   // F charge
  Double_t m1      = bases->f1Ptr->GetMass  ();   // F mass
  Double_t m2      = m1;                          // F mass
  Int_t    hela    = bases->fHelFinal[0];         // f1 helicity
  Int_t    hel1    = bases->fHelFinal[1];         // f1 helicity
  Int_t    hel2    = bases->fHelFinal[2];         // f2 helicity
  Double_t color   = bases->f1Ptr->GetColor();    // color factor
  Int_t    islev   = color > 1. ? 101 : 0;  	  // shower level
  Int_t    icf     = 1;                           // color flux id
  Double_t rq2z    = pv[1].Mag();

#if 0
//#ifdef __DEBUG__
  cerr << " -------------------------- " << endl;
  cerr << " 1 pid=" << idx << " m=" << 0.  << " Q=" << 0.
       << " pv=(" << (*qp[0])(0) << ", "
                  << (*qp[0])(1) << ", "
                  << (*qp[0])(2) << ", "
                  << (*qp[0])(3) << ") " << endl;
  cerr << " 2 pid=" << idz << " m=" << rq2z << " Q=" << 0.
       << " pv=(" << (*qp[1])(0) << ", "
                  << (*qp[1])(1) << ", "
                  << (*qp[1])(2) << ", "
                  << (*qp[1])(3) << ") " << endl;
  cerr << " 3 pid=" << idf << " m=" << m1 << " Q=" << chrg
       << " pv=(" << (*qp[2])(0) << ", "
                  << (*qp[2])(1) << ", "
                  << (*qp[2])(2) << ", "
                  << (*qp[2])(3) << ") " << endl;
  cerr << " 4 pid=" <<-idf << " m=" << m2 << " Q=" << -chrg
       << " pv=(" << (*qp[3])(0) << ", "
                  << (*qp[3])(1) << ", "
                  << (*qp[3])(2) << ", "
                  << (*qp[3])(3) << ") " << endl;
  TVector qcm(4), qz(4);
  qz  = *qp[2] + *qp[3];
  qcm = *qp[0] +qz;
  cerr << " pz=(" << qz[0] << ", "
                  << qz[1] << ", "
                  << qz[2] << ", "
                  << qz[3] << ") " << endl;
  cerr << " pcm=(" << qcm[0] << ", "
                   << qcm[1] << ", "
                   << qcm[2] << ", "
                   << qcm[3] << ") " << endl;
#endif

  //                              No. PID  Mass  Charge   pv   Nd 1st Mom  hel  col shower
  new (partons[0]) JSFSpringParton(1, ida,   0.,    0., *qp[0], 0, 0,  0, hela,   0,     0);
  new (partons[1]) JSFSpringParton(2, idz, rq2z,    0., *qp[1], 2, 3,  0,    0,   0,     0);
  new (partons[2]) JSFSpringParton(3, idf,   m1,  chrg, *qp[4], 0, 0,  2, hel1, icf, islev);
  new (partons[3]) JSFSpringParton(4,-idf,   m2, -chrg, *qp[5], 0, 0,  2, hel2, icf, islev);
  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class AZBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
AZBases::AZBases(const char *name, const char *title)
         : JSFBases   (name, title), 
#if 0
           fLambda    (1000.),
           fC0        (0.),
           fC1        (1.),
           fC2        (1.),
           fC3        (1.),
           fMass      ( 120.),
#endif
           fEcmInit   (1000.),
           fISR       ( 1),
           fBeamStr   ( 1),
           fBeamWidth (0.002),
           fPole      (0.),
           fZModesLo  ( 1),
           fZModesHi  (12),
           fZBosonPtr ( 0),
           fPhotonPtr ( 0),
           fZBoost    (0.),
           fEcmIP     (fEcmInit),
           fQ2AZ      (0.),
           fQ2Z       (0.),
           fZModePtr  (0),
           f1Ptr      (0),
           f2Ptr      (0),
           fCosTheta  (0.),
           fPhi       (0.),
           fCosThetaF (0.),
           fPhiF      (0.),
           fR_BW_m    (0),
           fR_BW_p    (0),
           fR_BS_m    (0),
           fR_BS_p    (0),
           fR_ISR_var (0),
           fR_ISR_side(0)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init azbases " << endl;
  
  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("AZBases.CosthRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.PhiOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.CosthFRange","-1.0 1.0"));
  ins >> fXL[2] >> fXU[2];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.PhiFOverPiRange","0.0 2.0"));
  ins >> fXL[3] >> fXU[3];

#if 0
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.Lambda","1000."));    // Lambda [GeV]
  ins >> fLambda;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.C0","0.")); 	  	 // C_0
  ins >> fC0;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.C1","1.")); 		 // C_1
  ins >> fC1;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.C2","1.")); 		 // C_2
  ins >> fC2;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.C3","1.")); 		 // C_3
  ins >> fC3;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.MassX","120.")); 	 // M_x [GeV]
  ins >> fMass;
#endif

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.Ecm","1000."));       // E_cm (1TeV)
  ins >> fEcmInit;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.Beamstrahlung","1")); // BmStr (on)
  ins >> fBeamStr;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.BeamWidth","0.002")); // Beam energy spread
  ins >> fBeamWidth;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.Bremsstrahlung","1"));// ISR (on)
  ins >> fISR;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.Pole","0."));         // electron polarization
  ins >> fPole;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.ZModesLo","1"));      // Z decay mode lo
  ins >> fZModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("AZBases.ZModesHi","12"));     // Z decay mode hi
  ins >> fZModesHi;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombInitial, 0., 1., 0, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 0, 1);
  DefineVariable(fZDecayMode    , 0., 1., 0, 1);
  DefineVariable(fXQ2Z          , 0., 1., 0, 1);
  //--
  //  cos(theta) and phi
  //--
  fXL[1] = fXL[1]*TMath::Pi();
  fXU[1] = fXU[1]*TMath::Pi();
  fXL[3] = fXL[3]*TMath::Pi();
  fXU[3] = fXU[3]*TMath::Pi();

  DefineVariable(fCosTheta , fXL[0], fXU[0], 1, 1);
  DefineVariable(fPhi      , fXL[1], fXU[1], 0, 1);
  DefineVariable(fCosThetaF, fXL[2], fXU[2], 0, 1);
  DefineVariable(fPhiF     , fXL[3], fXU[3], 0, 1);

  //--
  //  Beamstrahlung and natural Ebeam spread
  //--
  if (fBeamStr == 1) {
    DefineVariable(fR_BW_m, 0., 1., 0, 1);
    DefineVariable(fR_BW_p, 0., 1., 0, 1);
    DefineVariable(fR_BS_m, 0., 1., 0, 1);
    DefineVariable(fR_BS_p, 0., 1., 0, 1);
  }
 
  //--
  //  ISR
  //--
  if (fISR==1) {
    DefineVariable(fR_ISR_var,  0., 1., 0, 1);
    DefineVariable(fR_ISR_side, 0., 1., 0, 1);
  }

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(20000);

  SetTuneValue (1.5);
  SetIteration1(0.05, 20);
  SetIteration2(0.05, 50);

}
// --------------------------
//  D-tor
// --------------------------
AZBases::~AZBases()
{
  delete fZBosonPtr;
  delete fPhotonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t AZBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.; // Jacobian factor
  Double_t eminus;        // E_e- after ISR and beamstrahlung
  Double_t eplus;         // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  if (fBeamStr == 1) {
    bsWeight = fBM->GetWeight(fR_BW_m, fR_BW_p,
                              fR_BS_m, fR_BS_p,
                              eminus, eplus);
#if 0
    eplus   *= fBM->GetNominalEnergy();
    eminus  *= fBM->GetNominalEnergy();
#else
    eplus   *= fEcmInit/2.;
    eminus  *= fEcmInit/2.;
#endif
    fEcmIP   = 2.*TMath::Sqrt(eplus*eminus);     // ignore electron mass 
  } else {
    fEcmIP   = fEcmInit;      
    eplus    = fEcmInit/2.;
    eminus   = eplus;
  }
   
  // --------------------------------------------
  //  Initial State Radiation
  // --------------------------------------------

  static const Double_t kFisr = 2. * kAlpha0 / TMath::Pi();

  Double_t beta_isr = kFisr * (2.*(TMath::Log(fEcmIP/kM_e) - 1.));

  if (fISR == 1) {
    Double_t xisr    = TMath::Power(fR_ISR_var,1./beta_isr); // Ephoton / Ebeam 
             fEcmIP *= TMath::Sqrt(1. - xisr);                // reduced Ecm
    
    if (fR_ISR_side > 0.5) {
      eplus  *= 1. - xisr;
    } else {
      eminus *= 1. - xisr;
    }
  } // if Bremsstrahlung


  fZBoost = eminus - eplus; // P_z of the cm system after ISR and beamstrahlung

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  Double_t m0 = 0.; // photon

  GENDecayMode *fZModePtr = fZBosonPtr->PickMode(fZDecayMode, weight, fZMode);
  bsWeight *= weight;
  f1Ptr = static_cast<GENPDTEntry *>(fZModePtr->At(0));
  f2Ptr = static_cast<GENPDTEntry *>(fZModePtr->At(1));
  Double_t m1   = f1Ptr->GetMass();
  Double_t m2   = f2Ptr->GetMass();

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fEcmIP < m0 + m1 + m2) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for e- is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------
  fQ2AZ = fEcmIP*fEcmIP;

  Double_t rs   = fEcmIP;
  Double_t qmin = m1 + m2;
  Double_t qmax = rs - m0;
#ifndef __ZEROWIDTH__
  fQ2Z = fZBosonPtr->GetQ2BW(qmin, qmax, fXQ2Z, weight);
#else
  fQ2Z = TMath::Power(fZBosonPtr->GetMass(),2);
  weight = kPi*fZBosonPtr->GetMass()*fZBosonPtr->GetWidth();
#endif
  bsWeight *= weight;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------

  GENBranch zbranch (fQ2Z , fCosThetaF, fPhiF, m1*m1, m2*m2);
  GENBranch cmbranch(fQ2AZ, fCosTheta , fPhi , &zbranch, m0*m0);

  Double_t sigma = DSigmaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------

  Xh_fill( 1, fEcmIP           , (bsWeight*sigma));
  Xh_fill( 2, fCosTheta        , (bsWeight*sigma));
  Xh_fill( 3, fPhi             , (bsWeight*sigma));
  Xh_fill( 4, TMath::Sqrt(fQ2Z), (bsWeight*sigma));
  Xh_fill( 5, fCosThetaF       , (bsWeight*sigma));
  Xh_fill( 6, fPhiF            , (bsWeight*sigma));
  Xh_fill( 7, (Double_t)fJCombI, (bsWeight*sigma));
  Xh_fill( 8, (Double_t)fJCombF, (bsWeight*sigma));
  Xh_fill( 9, (Double_t)fZMode , (bsWeight*sigma));

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t AZBases::DSigmaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  Double_t q2z  = cmbranch.GetM12();
  Double_t q2a  = cmbranch.GetM22();
  Double_t cosz = cmbranch.GetCosTheta();
  Double_t phiz = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2z, q2a, cmframe, cosz, phiz, 0);
  ANL4DVector pz = phaseCM.GetFourMomentum(0);
  fP[0] = phaseCM.GetFourMomentum(1);
  Double_t betaz = phaseCM.GetBetaBar();
  if (betaz <= 0.) return 0.;

  GENBranch &zbranch = * cmbranch.GetBranchPtr(0);
  Double_t cosf = zbranch.GetCosTheta();
  Double_t phif = zbranch.GetPhi     ();
  Double_t m12  = zbranch.GetM12();
  Double_t m22  = zbranch.GetM22();
  GENPhase2 phaseZ(pz, m12, m22, cmframe, cosf, phif, 1);
  fP[1] = phaseZ.GetFourMomentum(0);
  fP[2] = phaseZ.GetFourMomentum(1);
  fM[1] = TMath::Sqrt(m12);
  fM[2] = TMath::Sqrt(m22);
  Double_t betaf = phaseZ.GetBetaBar();
  if (betaf <= 0.) return 0.;

  Double_t eb     = rs/2.;
  Double_t pb     = TMath::Sqrt((eb-kM_e)*(eb+kM_e));
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, eb);
#if 0
  cerr << "fP[0]=(" << fP[0](0) << ", " << fP[0](1) << ", " << fP[0](2) << ", " << fP[0](3) << ")" << endl; 
  cerr << "fP[1]=(" << fP[1](0) << ", " << fP[1](1) << ", " << fP[1](2) << ", " << fP[1](3) << ")" << endl; 
  cerr << "fP[2]=(" << fP[2](0) << ", " << fP[2](1) << ", " << fP[2](2) << ", " << fP[2](3) << ")" << endl; 
  ANL4DVector qvz = fP[1]+fP[2];
  ANL4DVector qvf = fP[0]+qvz;
  cerr << "qvf=(" << qvf(0) << ", " << qvf(1) << ", " << qvf(2) << ", " << qvf(3) << ")" << endl; 
#endif
  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared(cmbranch);

  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbr  = 2;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));

  Double_t identp = 1.;                    // identical particle factor
  Double_t dPhase = kFact * betaz * betaf; // phase space factor
  Double_t flux   = 1./(2.* s * beta_e);   // beam flux factor
  Double_t spin   = 1./2.;                 // spin average for e+

  Double_t sigma  = identp * flux * spin * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                              // now in [fb]

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t AZBases::AmpSquared(GENBranch &cmbranch)
{
  Double_t  color = f1Ptr->GetColor();
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2) * color;

#ifdef __NODECAY__
  Double_t q2z    = cmbranch.GetM22();
  Double_t rq2z   = TMath::Sqrt(q2z);
  Double_t gamz   = fZBosonPtr->GetWidth();
           amp2  *= (2*kM_z*gamz)/(TMath::Power((rq2z - kM_z)*(rq2z + kM_z),2)
			                          + TMath::Power(kM_z*gamz,2));
  amp2 *= k8Pi; // branch factor correction
#endif
  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t AZBases::FullAmplitude()
{
   Double_t gamz   = fZBosonPtr->GetWidth();

   Double_t qf     = f1Ptr->GetCharge();
   Double_t t3f    = f1Ptr->GetISpin();
   Double_t glz    = -kGz*(t3f - qf*kSin2W);
   Double_t grz    = -kGz*(    - qf*kSin2W);

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELFermion ep(fK[1], kM_e, fHelInitial[1], -1, kIsOutgoing);

   HELVector  af(fP[0], 0., fHelFinal[0], +1); // photon

   HELFermion f (fP[1], fM[1], fHelFinal [1], +1, kIsOutgoing);
   HELFermion fb(fP[2], fM[2], fHelFinal [2], -1, kIsIncoming);
   HELVector  zf(fb, f, glz, grz, kM_z, gamz);

   Complex_t amp = AmpEEtoAZ(em, ep, af, zf);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEEtoAZ()
// --------------------------
Complex_t AZBases::AmpEEtoAZ(const HELFermion &em,
                             const HELFermion &ep,
                             const HELVector  &af,
                             const HELVector  &zf)
{
   //-------------------
   // Coupling consts.
   //-------------------
#if 0
   Double_t  a1    = fC1*kCos2W + fC2*kSin2W;
   Double_t  a2    = (fC1 - fC2) * kSinCosW;
   Double_t  a3    = fC1*kSin2W + fC2*kCos2W;
   Double_t  gaax  = a1*kSqh/fLambda;
   Double_t  gazx  = a2*kSqh/fLambda;
   Double_t  gzzx  = a3*kSqh/fLambda;
#if 1
   static Int_t ncall = 0;
   if (!ncall) {
      ncall = 1;
      cerr << " a1 = " << a1
           << " a2 = " << a2
           << " a3 = " << a3 << endl;
      cerr << " kSqh = "    << kSqh 
           << " fLambda = " << fLambda << endl;
   }
#endif
#endif

   Double_t  qe    = -1.;
   Double_t  ge    = -qe*kGe;
   Double_t  gal   = -ge;

   Double_t  t3e   = -1./2.;
   Double_t  glze  = -kGz*(t3e - qe*kSin2W);
   Double_t  grze  = -kGz*(    - qe*kSin2W);

   Double_t  gamz  = fZBosonPtr->GetWidth();
   Double_t  me    = kM_e;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
  // AZ Production Amplitude
   //---------------------------
   //--
   // A from e-
   //--
   //   HELFermion(Double_t     eb,    // energy       of beam e+/-  [GeV]
   //              Double_t     ea,    // energy       of ISR photon [GeV]
   //	           Double_t     shlf,  // sin(theta/2) of ISR photon
   //	           Double_t     chlf,  // cos(theta/2) of ISR photon
   //	           Double_t     phi,   // phi          of ISR photon
   //	           Double_t     gal,   // QED coupling of e+/-
   //	           Double_t     me,    // electron mass
   //	           Int_t        nhe,   // helicity (-1,+1) of e+/-
   //              Int_t        nha,   // helicity (-1,+1) of ISR photon
   //	           Bool_t       isincom = kFALSE);
   //
   ANL4DVector pem  = em.GetFourMomentum();
   ANL4DVector pa   = af.GetFourMomentum();
   Double_t    eb   = pem(0);
   Double_t    ea   = pa (0);
   Double_t    cs   = pa.CosTheta();
   Double_t    phi  = pa.Phi();
   Double_t    chlf = TMath::Sqrt((1.+cs)/2);
   Double_t    shlf = TMath::Sqrt((1.-cs)/2);
   Int_t       nhe  = em.GetHelicity();
   Int_t       nhp  = ep.GetHelicity();
   Int_t       nha  = af.GetHelicity();

   //--
   // photon from e-
   //--
   static const Bool_t kIsIncoming = kTRUE;
   HELFermion ein(eb, ea, shlf, chlf, phi, nhe, nha, kIsIncoming, gal, me); 
   //HELFermion ein(em, af, gal, gal, me, 0.); 
   Complex_t  ampaz = HELVertex(ein, ep, zf, glze, grze);

   //--
   // photon from e+
   //--
   static const Bool_t kIsOutgoing = kFALSE;
   HELFermion eot(eb, ea, shlf, chlf, phi, nhp, nha, kIsOutgoing, gal, me); 
   //HELFermion eot(ep, af, gal, gal, me, 0.); 
   Complex_t  ampza = HELVertex(em, eot, zf, glze, grze);

   //--
   // Sum of the two
   //--
   Complex_t amp = ampaz + ampza;
#endif /* end __PHASESPACE__ */

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void AZBases::Userin()
{
  // --------------------------------------------
  //  Open beamstrahlung data
  // --------------------------------------------
  string bsfiledir;
  string bsfilename;

  if (fBeamStr) {
    stringstream ins(gJSF->Env()->GetValue("AZBases.BeamstrahlungFilepath",
                                  "/proj/soft/jsf/pro/data/bsdata/"));
    ins >> bsfiledir;

    ins.clear();
    ins.str(gJSF->Env()->GetValue("AZBases.BeamstrahlungFilename","trc500"));
    ins >> bsfilename;
    
    TFile *fBeamFile = new TFile((bsfiledir+bsfilename+".root").data());
    if (!fBeamFile) {
      cerr << " Unable to open a file for beamstrahlung" << endl;
      abort();
    }
    
  // --------------------------------------------
  //  Initialize beam generator
  // --------------------------------------------

    fBM = (JSFBeamGenerationCain*)fBeamFile->Get(bsfilename.data());
    fBM->SetIBParameters(fBeamWidth);

    fBM->MakeBSMap();
    fBM->Print(); 

    cout << " Nominal Energy (beamstrahlung) = " << fBM->GetNominalEnergy() 
         << " beam width     (beamstrahlung) ="  << fBM->GetIBWidth() 
         << endl;

    if (!(fBM->GetNominalEnergy() == (fEcmInit/2))) {
      cout << "Nominal energy from beamstrahung is " << fBM->GetNominalEnergy() 
           << " which is different from fEcm/2: "    << (fEcmInit/2)
           << endl;        
    } // check the energy homogeneity
  } // if beamstrahlung is on

  // --------------------------------------------
  //  Initialize Z decay table
  // --------------------------------------------
  if (!fZBosonPtr) fZBosonPtr = new GENPDTZBoson();
  for (Int_t m=1; m<=fZBosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fZBosonPtr->GetMode(m); 
     if (mp && (m<fZModesLo || m>fZModesHi)) {
        mp->Lock();
     }
  }
  fZBosonPtr->DebugPrint();

  if (!fPhotonPtr) fPhotonPtr = new GENPDTPhoton();
  //fPhotonPtr->DebugPrint();

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Xh_init( 1,     0., fEcmInit*1.1, 50, "Ecm"   );
  Xh_init( 2, fXL[0], fXU[0],       50, "Costh" );
  Xh_init( 3, fXL[1], fXU[1],       50, "Phi"   );
  Xh_init( 4,    70.,   110.,       50, "Mz"    );
  Xh_init( 5, fXL[2], fXU[2],       50, "CosthF");
  Xh_init( 6, fXL[3], fXU[3],       50, "PhiF"  );
  Xh_init( 7,     0.,     2.,        2, "Helin ");
  Xh_init( 8,     0.,     4.,        2, "Helot ");
  Xh_init( 9,     0.,    12.,       12, "Z mode");
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void AZBases::Userout()
{
  cout << "End of AZBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcmInit << " [GeV]   "    << endl
       << "Beamstrahlung        = " << (fBeamStr ? "on" : "off")  << endl
       << "Bremsstrahlung       = " << (fISR     ? "on" : "off")  << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"  << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void AZBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 2;
   static const Int_t kIHelComb[kNi][2] = {{-1, +1},
                                           {+1, -1}};
   static const Int_t kNf = 4;
   static const Int_t kFHelComb[kNf][3] = {{-1, -1, +1},
                                           {-1, +1, -1},
                                           {+1, -1, +1},
                                           {+1, +1, -1}};
   Double_t helm = (1. - fPole)/2.;
   if (fHelCombInitial < helm) {
      fJCombI = 0;
   } else {
      fJCombI = 1;
   }
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   weight = kNf;
}
