#ifndef AZSPRING_H
#define AZSPRING_H
//*****************************************************************************
//* =====================
//*  AZSpring
//* =====================
//*  
//* (Description)
//*    e+e- -> ZA generator
//*
//* (Update Record)
//*    2017/10/14  K.Fujii	Original version.
//*
//*****************************************************************************

#include "TNamed.h"
#include "JSFModule.h"
#include "JSFBases.h"
#include "JSFSpring.h"
#include "JSFBeamGeneration.h"

#include "HELLib.h"
#include "GENLib.h"

//_______________________________________________________________________
// =====================
//  class AZBases
// =====================
//-----------------------------------------------------------------------
class AZBases : public JSFBases {
friend class AZSpringBuf;
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  AZBases(const char *name  = "AZBases", 
          const char *title = "AZ Bases");
  virtual ~AZBases();

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetEcmInit  ()           const { return fEcmInit;   }
#if 0
  Double_t GetMass     ()           const { return fMass;      }
#endif

  Double_t GetQ2AZ     ()           const { return fQ2AZ;      }
  Double_t GetCosTheta ()           const { return fCosTheta;  }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetQ2Z      ()           const { return fQ2Z;       }
  Double_t GetCosThetaF()           const { return fCosThetaF; }
  Double_t GetPhiF     ()           const { return fPhiF;      }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  void     SetEcmInit  (Double_t ecm    ) { fEcmInit   = ecm;  }
#if 0
  void     SetMass     (Double_t m      ) { fMass      = m;    }
#endif
  void     SetISR      (Bool_t b = kTRUE) { fISR       = b;    }
  void     SetBeamStr  (Bool_t b = kTRUE) { fBeamStr   = b;    }
  void     SetBeamWidth(Double_t w      ) { fBeamWidth = w;    }
  void     SetPole     (Double_t p      ) { fPole      = p;    }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual void     Userin();   // Bases user initialization
  virtual void     Userout();  // Bases user output 

  Double_t Func();     // Bases integration function.

  // ----------------------
  //   Utility methods
  // ----------------------
private:
  void     SelectHelicities(Double_t &weight);

  Double_t  DSigmaDX     (GENBranch &cmbranch);
  Double_t  AmpSquared   (GENBranch &cmbranch);
  Complex_t FullAmplitude();
  Complex_t AmpEEtoAZ    (const HELFermion &em,
                          const HELFermion &ep,
                          const HELVector  &af,
                          const HELVector  &zf);

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Job parameters
  // ----------------
#if 0
  Double_t fLambda;         // lambda
  Double_t fC0;             // C_0
  Double_t fC1;             // C_1
  Double_t fC2;             // C_2
  Double_t fC3;             // C_3

  Double_t fMass;           // m_x    : mass  of X
#endif

  Double_t fEcmInit;        // Initial Ecm
  Int_t    fISR;            // ISR on?
  Int_t    fBeamStr;        // Beamstrahlung on?
  Double_t fBeamWidth;      // Beam width relative to Ebm(nominal)
  Double_t fPole;           // electron polarization
  Int_t    fZModesLo;       // Z decay mode lo;
  Int_t    fZModesHi;       // Z decay mode hi;

  // ----------------
  //  Particle Data
  // ----------------
  GENPDTZBoson *fZBosonPtr;       //! PD table entry of "Z"
  GENPDTPhoton *fPhotonPtr;       //! PD table entry of "Gamma"

  // ----------------
  //  Event info
  // ----------------
  Double_t       fZBoost;         // p_z(cm)      in lab frame
  Double_t       fEcmIP;          // Ecm after B-strahlung
  Double_t       fQ2AZ;           // q^2 of AZ system
  Double_t       fQ2Z;            // q^2 of final state Z
  GENDecayMode  *fZModePtr;       // pointer to Z decay mode
  GENPDTEntry   *f1Ptr;           // point to 1st Z daughter
  GENPDTEntry   *f2Ptr;           // point to 2nd Z daughter
  Int_t          fHelInitial[2];  // initial state helicities
  Int_t          fHelFinal  [3];  // final   state helicities
  Int_t          fJCombI;         // intial helicity combination
  Int_t          fJCombF;         // final  helicity combination
  Int_t          fZMode;          // Z decay mode
  ANL4DVector    fK[2];           // [0,1] = [e-, e+]
  ANL4DVector    fP[3];           // [0,1,2] = [a, fz1, fz2]
  Double_t       fM[3];           // [0,1,2] = [0, m1 , m2 ]

  // ----------------
  //  Integ. vars
  // ----------------
  Double_t       fHelCombInitial; // initial state helicity combination
  Double_t       fHelCombFinal;   // final   state helicity combination
  Double_t       fXDecayMode;     // decay mode selector for X
  Double_t       fZDecayMode;     // decay mode selector for Z
  Double_t       fCosTheta;       // cos(theta_z) in cm  frame
  Double_t       fPhi;            // phi_z        in cm  frame
  Double_t       fXQ2Z;           // q^2 of final state Z
  Double_t       fCosThetaF;      // cos(theta_f) in Z   frame
  Double_t       fPhiF;           // phi_f        in Z   frame

  Double_t       fR_BW_m;         //! naturl Ebeam spread for e-
  Double_t       fR_BW_p;         //!                     for e+
  Double_t       fR_BS_m;         //! beamstrahlung       for e-
  Double_t       fR_BS_p;         //!                     for e+

  Double_t       fR_ISR_var;      //! initial state radiation
  Double_t       fR_ISR_side;     //! from which side ? (e-/e+) 

  Double_t       fXU[4];          //! [0,1,2,3] = (cosz, phiz,
  Double_t       fXL[4];          //!              cosf, phif)

  TFile   *fBeamFile;         //! Beamstrahlung data file
  JSFBeamGenerationCain *fBM; //! Beamstrahlung data

  ClassDef(AZBases, 1) // Bases for e+e- -> XX process
};

class AZSpring;

//_______________________________________________________________________
// =====================
//  class AZSpringBuf
// =====================
//-----------------------------------------------------------------------
class AZSpringBuf : public JSFSpringBuf {
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  AZSpringBuf(const char *name   = "AZSpringBuf", 
              const char *title  = "AZ Spring test event buffer",
	      AZSpring   *spring = 0)
	   : JSFSpringBuf(name,title,(JSFSpring*)spring) {} 
  virtual ~AZSpringBuf() {}

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetQ2AZ     ()           const { return fQ2AZ;      }
  Double_t GetCosTheta ()           const { return fCosTheta;  }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetQ2Z      ()           const { return fQ2Z;       }
  Double_t GetCosThetaF()           const { return fCosThetaF; }
  Double_t GetPhiF     ()           const { return fPhiF;      }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual Bool_t SetPartons();

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Event info
  // ----------------
  Double_t fQ2AZ;           // q^2 of AZ system
  Double_t fCosTheta;       // cos(theta_x) in cm  frame
  Double_t fPhi;            // phi_x        in cm  frame
  Double_t fQ2Z;            // q^2 of final state Z
  Double_t fCosThetaF;      // cos(theta_f) in Z   frame
  Double_t fPhiF;           // phi_f        in Z   frame
  Double_t fZBoost;         // p_z(cm)      in lab frame
  Double_t fEcmIP;          // Ecm after B-strahlung

  ClassDef(AZSpringBuf, 1)  // AZSpring event buffer
};

//_______________________________________________________________________
// =====================
//  class AZSpring
// =====================
//-----------------------------------------------------------------------
class AZSpring : public JSFSpring {
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
public:
   AZSpring(const char *name  = "AZSpring", 
            const char *title = "AZ Spring test",
            AZBases    *bases = 0);
   virtual ~AZSpring();

   ClassDef(AZSpring, 1)  // AZSpring class
};
#endif
