#ifndef __S4CALCLUSTER__
#define __S4CALCLUSTER__

//*************************************************************************
//* ---------------------
//* S4CALCluster Class
//* ---------------------
//* (Description)
//*	represent calorimeter cluster 
//* (Requires)
//*	S4CALHit
//* (Provides)
//*	S4CALCluster		: made object class
//* (Update Record)
//*	2004/09/03  S.Yamamoto	Original version.
//*	2004/11/04  K.Fujii	Added drawable attribute.
//*     2008/08/04  A.Miyamoto  Added theta/Phi dependant calibration 
//*************************************************************************

#include "TObjArray.h"
#include "TVector3.h"
#include "S4CALHit.h"
#include "S4CALTypeDef.h"
#include "S4HYBTrack.h"
#include "TAttDrawable.h"

class S4CALCluster : public TObjArray, public TAttLockable, public TAttDrawable {

public:
 S4CALCluster(S4HYBTrack      *trackp       = 0, 
              EClusterType_t   clusterype   = kgCheated_cl, 
              EParticleType_t  particletype = kgNotDecided_p);
 virtual ~S4CALCluster();
  
 virtual void 	DebugPrint(const Char_t *opt = "");

 using TAttDrawable::Draw;
 virtual void Draw(Int_t color, const Char_t *opt = "");
 
  // Data Getters
 virtual Int_t           GetCALType()       { if (!fDone) Update(); return fCALType;         }
 virtual Double_t        GetR()             { if (!fDone) Update(); return fClusterR;        }
 virtual Double_t        GetTheta()         { if (!fDone) Update(); return fClusterTheta;    }
 virtual Double_t        GetPhi()           { if (!fDone) Update(); return fClusterPhi;      }
 virtual Double_t        GetEnergy()        { if (!fDone) Update(); return fClusterEnergy;   }
 virtual TVector3        GetMomentum()      { if (!fDone) Update(); return fClusterMomentum; }
 virtual TObjArray      &GetTracks()        { if (!fDone) Update(); return fTracks;          }
 virtual Double_t        GetTime()          { if (!fDone) Update(); return fTime;            }
 virtual TVector3        GetPosition()      { if (!fDone) Update(); return fPosition; }
 virtual Double_t        GetFracEM()        { if (!fDone) Update(); return fFracEM;            }
 virtual EClusterType_t  GetClusterType()   { if (!fDone) Update(); return fClusterType;     }
 virtual EParticleType_t GetParticleType()  { if (!fDone) Update(); return fParticleType;    }   

 // hit cut getters & setters
 static  Double_t    GetEhitCut()                  { return fgHitEmin;    }
 static  Double_t    GetTOFhitCut()                { return fgHitTOFmax;  } 
 static  Double_t    GetDThitCut()                 { return fgHitDTmax;   } 
 static  Double_t    GetDXhitCut()                 { return fgHitDXmax;   } 
 static  Double_t    GetEMCut()                    { return fgEMCut;      } 
 static  Double_t    GetCalibFactorForEM()         { return fgCalibFactorForEM;} 
 static  Double_t    GetCalibFactorForHD()         { return fgCalibFactorForHD;} 
 static  Double_t    GetCalibBERatioForEM()        { return fgCalibBERatioForEM;}
 static  Double_t    GetCalibBERatioForHD()        { return fgCalibBERatioForHD;}

 static  void        SetEhitCut(Double_t value)    { fgHitEmin   = value; }
 static  void        SetTOFhitCut(Double_t value)  { fgHitTOFmax = value; }
 static  void        SetDThitCut (Double_t value)  { fgHitDTmax  = value; }
 static  void        SetDXhitCut (Double_t value)  { fgHitDXmax  = value; }
 static  void        SetEMCut (Double_t value)     { fgEMCut     = value; }
 static  void        SetCalibFactorForEM (Double_t value)  { fgCalibFactorForEM  = value; }
 static  void        SetCalibFactorForHD (Double_t value)  { fgCalibFactorForHD  = value; }


 static  void        SetCalibBERatioForEM(Double_t value) { fgCalibBERatioForEM=value;}
 static  void        SetCalibBERatioForHD(Double_t value) { fgCalibBERatioForHD=value;}

 static  void        PrintConstants();

 // Service method
 virtual void        Add(const S4CALHit *hit);                 // Add hit
 virtual Bool_t      Append(const S4CALHit *hit);              // Append hit
 virtual S4HYBTrack *GetTrackPtr(Int_t i=0);                   // Get i-th track pointer
 virtual Bool_t      IsOK(S4CALHit &hit);                      // Cut hit

private: 
 virtual void      Add(TObject* hit) { TObjArray::Add(hit); }  // Add hit
 virtual void      Update();                                   // Update cluster information
 virtual void      Calibrate();                                // energy calibration
  
private:
 Int_t           fCALType;          // CalType of Cluster
 Double_t        fClusterR;         // Cluster Position(cm)
 Double_t        fClusterTheta;     // Cluster Position(radian)
 Double_t        fClusterPhi;       // Cluster Position(radian)
 Double_t        fClusterEnergy;    // Energy of Cluster(GeV)
 TVector3        fClusterMomentum;  // Momentum of Cluster(GeV)
 TObjArray       fTracks;           // Cluster maching track array  
 Double_t        fTime;             // hit time at Cell(nsec)
 TVector3        fPosition;         // Cluster Position x(cm),y(cm),z(cm)
 Double_t        fFracEM;           // Energy fraction in EM
 EClusterType_t  fClusterType;      // Cluster Type
 EParticleType_t fParticleType;     // Particle Type
 Bool_t          fDone;	            // (1,0) = (clustering done, not yet)

 static Double_t fgHitEmin;         //! Hit energy cut 
 static Double_t fgHitTOFmax;       //! Hit tof cut 
 static Double_t fgHitDTmax;        //! Hit time cut
 static Double_t fgHitDXmax;        //! Hit position cut
 static Double_t fgEMCut;           //! cut for CalibrationFactor
 static Double_t fgCalibFactorForEM;//! CalibrationFactor for EM
 static Double_t fgCalibFactorForHD;//! CalibrationFactor for HD

 static Double_t fgCalibBERatioForEM;//! Calibration Factor, Barrel/EM ratio for EM
 static Double_t fgCalibBERatioForHD;//! Calibration Factor, Barrel/EM ratio for HD
//        if these constants are negative, no correction is applied.


 ClassDef(S4CALCluster, 2) // Cluster data information
};

#endif
