//*************************************************************************   
//* ---------------------                      
//* S4VCluster Class
//* ---------------------   
//* (Description) 
//*     represent calorimeter cluster
//* (Requires)
//*     S4CALHit
//* (Provides)  
//*     S4VCluster            : made object class
//* (Update Record)
//*     2004/09/03  S.Yamamoto  Original version.
//*     2004/11/15  S.Yamamoto  Add IsOK
//*************************************************************************

#include "S4VCluster.h"
#include "S4CALExactHit.h"
#include "S4CALPreHit.h"
#include "S4TPCPostHit.h"
#include "S4CALAddress.h"
#include "JSFSteer.h"

#include "TVirtualPad.h"
#include "TPolyMarker3D.h"
#include <iostream>

//#define __DEBUG__

//Double_t S4VCluster::fgCalibFactorForEM = 30. * 1.73 / 1.01 / 1.41; // Yamamoto
//Double_t S4VCluster::fgCalibFactorForHD = 30. * 1.30 / 1.07 / 1.06; // Yamamoto
Double_t S4VCluster::fgCalibFactorForEM = 24.1237; // Yamamoto              
Double_t S4VCluster::fgCalibFactorForHD = 29.9906; // Yamamoto         

ClassImp(S4VCluster)

//=====================================================================
//* constructor -------------------------------------------------------
S4VCluster::S4VCluster()
           : TObjArray(500), 
	     fCALType(kgNotDecided),  
	     fEnergy(0.), fEnergyDeposit(0.),
	     fPosition(TVector3(0.,0.,0.)),
	     fMomentum(TVector3(0.,0.,0.)),
	     fThrustVector(TVector3(0.,0.,0.)),
	     fThrust(0.),
	     fBroadening(0.),
	     fTOF(0.),
	     fNHits(),
	     fDone(kFALSE),
	     fMeanLayerID(-100.),
	     fMinLayerID(1000),fMaxLayerID(-1000),
	     fClusterID(-1),
	     fMinPos(0.,0.,0.),
	     fMaxPos(0.,0.,0.)
{
  fCALHitsVector.clear();
  fPIDMap.clear();
  fPostHitIDMap.clear(); 
}

//=====================================================================
//* destructor --------------------------------------------------------

S4VCluster::~S4VCluster()
{
}

//=====================================================================
//* Add ---------------------------------------------------------------

void S4VCluster::Add(const S4CALHit *hitp)
{
   if (fDone) fDone = kFALSE;
   TObjArray::Add(const_cast<S4CALHit *>(hitp));
   fCALType |= (ECALType_t)hitp->GetCALType(); // set CALType
}

//=====================================================================
//* Calibrate ---------------------------------------------------------  
                                          
void S4VCluster::Calibrate()
{

  fEnergy = 0.;

  for(Int_t k = 0; k < (Int_t) fCALHitsVector.size(); k++){
    Bool_t isem = fCALHitsVector[k]->IsEM();
    Double_t calibFactor = isem ? fgCalibFactorForEM : fgCalibFactorForHD;
    fEnergy += calibFactor*(fCALHitsVector[k]->GetEnergyDeposit());
  }

}

//=====================================================================
//* MakeThrust --------------------------------------------------------  

void S4VCluster::MakeThrust()
{
  
  // temporary!

  if(GetNHits() < 4){
    fThrustVector = TVector3(0.,0.,0.);
    fThrust       = -1;
    fBroadening   = -1;
    return;
  }

  TVector3 clpos(0.,0.,0.);      // center position of the cluster
  TVector3 hitpos(0.,0.,0.);     // hit cell position
  TVector3 Thrust(0.,0.,0.);
  TVector3 ThrustOrg(0.,0.,0.);  // unit vector of IP to center
  TVector3 ThrustOrth(0.,0.,0.); // orthogonal vector of ThrustOrg
  TVector3 ThrustMax(0.,0.,0.);  // temporary thrust
  Double_t dist    = 0.;         // T = sum|Pi*n|/sum|Pi| (Pi,n: pos. vector)
  Double_t dist2   = 0.;         // B = sum|Pi cross n|/sum|Pi| (Pi,n: vector)
  Double_t maxdist = 0;          // temporary maximum T
  Double_t mindist = 0;          // temporary B
  Double_t norm    = 0;          // normarization factor
  Int_t    LOOP    = 10;         // sampling number
  //Int_t    LOOP    = 50;          // sampling number

  std::vector<S4CALHit*> vhits = GetCALHitsVector();
  clpos = GetPosition();
  ThrustOrg = TVector3(clpos.X()/clpos.Mag(),
		       clpos.Y()/clpos.Mag(),
		       clpos.Z()/clpos.Mag());
  Thrust = ThrustOrg;
  ThrustOrth = ThrustOrg.Orthogonal();

  for (Int_t i = 0; i <= LOOP/2; i++){    // loop for theta
    //for (Int_t i = 0; i <= LOOP; i++){    // loop for theta

    Thrust = ThrustOrg;
    Thrust.Rotate(i*TMath::Pi()/LOOP, ThrustOrth);
    TVector3 ThrustTmp = Thrust;

    for (Int_t j = 1; j <= 2*LOOP; j++){  // loop for phi

      dist  = 0.;
      dist2 = 0.;
      norm  = 0.;

      Thrust = ThrustTmp;
      Thrust.Rotate(j*TMath::Pi()/LOOP, ThrustOrg);

      for (Int_t k = 0; k < (Int_t)vhits.size(); k++){
	hitpos = vhits[k] -> GetCellPosition();
	dist  += TMath::Abs( (hitpos-clpos)*(vhits[k]->GetEnergyDeposit())*Thrust );
	dist2 += ( ((hitpos-clpos)*(vhits[k]->GetEnergyDeposit())).Cross(Thrust) ).Mag();
	norm += ( (hitpos - clpos)*(vhits[k]->GetEnergyDeposit()) ).Mag();
      }

      if ( dist/norm > maxdist ){
	maxdist = dist/norm;
	mindist = dist2/norm;
	ThrustMax = Thrust;
      }
	  
    } // loop-j
  } // loop-i  

  fThrustVector = ThrustMax;
  fThrust       = maxdist;
  fBroadening   = mindist;

}

//=====================================================================
//* Update ------------------------------------------------------------

void S4VCluster::Update()
{

  //  static const Int_t emnlayers = gJSF->Env()->GetValue("J4CAL.EM.NLayers",38);
  static const Int_t emnlayers = S4CALAddress::GetEMNLayers();

  if(fDone) return;
  if(TObjArray::GetEntries() == 0) return;
  
  fCALHitsVector.clear(); // don't forget!

  TIter nexthit(this);
  
  Int_t    nhits   = 0;
  Int_t    layerID = 0;
  Double_t mlayer  = 0.;
  Double_t ecl     = 0.;
  Double_t tcl     = 0.;
  TVector3  xcl    = TVector3(0.,0.,0.);
  S4CALHit *hitp   = 0;
  
  fMinLayerID = 1000;
  fMaxLayerID = -1000;
  
  while ((hitp = dynamic_cast<S4CALHit *>(nexthit()))) {
    
    Double_t edep = hitp->GetEnergyDeposit();
    
    ecl += edep;
    xcl += edep * hitp->GetCellPosition(); 
    
    layerID = S4CALAddress::GetLayerID(hitp->GetCellID(), hitp->IsEM());
    //    std::cout << "!!!!! " << " " << layerID << " " << !hitp->IsEM() << std::endl;    
    if(!hitp->IsEM()) layerID += emnlayers;

    mlayer += layerID*edep;

    if( layerID > fMaxLayerID )  fMaxLayerID = layerID;
    if( layerID < fMinLayerID )  fMinLayerID = layerID;

    tcl += (hitp->GetTOF())*edep;
    nhits++;
    fCALHitsVector.push_back(hitp);
    
  }  // end of while

  //==================================================== 
  // calculate maximum/minimum layer position    
  Double_t minenergy = 0.;
  Double_t maxenergy = 0.;
  TVector3 minpos(0.,0.,0.);
  TVector3 maxpos(0.,0.,0.);
  std::vector<S4CALHit*>::iterator hitr;
  
  for(hitr = fCALHitsVector.begin(); hitr != fCALHitsVector.end(); hitr++){
    
    layerID = S4CALAddress::GetLayerID((*hitr)->GetCellID(),(*hitr)->IsEM());
    if( !((*hitr)->IsEM()) ) layerID += emnlayers;
    
    if (layerID == fMinLayerID) {
      minpos    += (*hitr)->GetCellPosition() * (*hitr)->GetEnergyDeposit();
      minenergy += (*hitr)->GetEnergyDeposit();
    }
    if (layerID == fMaxLayerID) {
      maxpos    += (*hitr)->GetCellPosition() * (*hitr)->GetEnergyDeposit();
      maxenergy += (*hitr)->GetEnergyDeposit();
    }
  } // end of for(hitr = ...)                                    
  
  minpos *= 1./minenergy;
  maxpos *= 1./maxenergy;
  
  fMinPos = minpos;
  fMaxPos = maxpos;
  //==================================================== 
  
  fEnergyDeposit = ecl;
  if (fEnergyDeposit <= 0.) return;
  fTOF       = tcl/ecl;
  fPosition  = xcl;
  fPosition *= 1./ecl;
  fNHits     = nhits;

  fMeanLayerID = mlayer/ecl;
  
  fDone = kTRUE;
  
  Calibrate();
  
  MakeThrust();
  
  SetPIDMap();
  
  // temporary momentum
  fMomentum = fPosition*fEnergy*(1./fPosition.Mag());
  
}

//=====================================================================
//* DebugPrint --------------------------------------------------------  

void S4VCluster::DebugPrint(const Char_t *opt)
{
}

//=====================================================================
//* Draw --------------------------------------------------------------  
                                          
void S4VCluster::Draw(Int_t color, const Char_t *opt)
{

  if (!gPad) return;

  TString option = opt;
  option.ToLower();
  gPad->cd();

  std::vector<S4CALHit*> hitv = GetCALHitsVector();

  for(Int_t i = 0; i < (Int_t) hitv.size(); i++) hitv[i] -> Draw(color,opt);

  gPad->Update();

}


////////////////////////////////////////////////////////////////////////
// Cheating method                                                    //

//=====================================================================
//* SetPIDMap ---------------------------------------------------------  

void S4VCluster::SetPIDMap()
{

  fPIDMap.clear();
  fPostHitIDMap.clear();

  TIter nexthit(this);
  S4CALHit* hitp = 0;
  while( (hitp = dynamic_cast<S4CALHit*>(nexthit())) ){

    Int_t pid    = 0;
    Int_t postid = 0;

    TObjArray *exhits = hitp->GetExactHitsPtr();
    TIter next(exhits);

    S4VExactHit *vexhit = 0;

    while( (vexhit = dynamic_cast<S4VExactHit*>(next())) ){

      S4CALExactHit *ehitp = dynamic_cast<S4CALExactHit*> (vexhit);
      S4CALPreHit *prehitp = ehitp -> GetPreHitPtr();

      if(prehitp -> GetPostHitPtr()){
	pid = prehitp->GetPostHitPtr()->GetPDGCode();
	postid = prehitp->GetPostHitPtr()->GetTrackID();
      }else{
	pid = prehitp->GetPDGCode();
	postid = prehitp->GetTrackID();
      }

      
      if( fPIDMap.count(pid) > 0 ){
	Double_t &edep = fPIDMap.find(pid)->second;
	edep += ehitp->GetEnergyDeposit();
      }else{
	fPIDMap[pid] = ehitp->GetEnergyDeposit();
      }

      if( fPostHitIDMap.count(postid) > 0 ){
	Double_t &edep = (fPostHitIDMap.find(postid)->second).second;
	edep += ehitp->GetEnergyDeposit();
      }else{
	fPostHitIDMap[postid] = std::pair<Int_t,Double_t>(pid,ehitp->GetEnergyDeposit());
      }

    } // next() loop

  } // nexthit() loop

}
