//*************************************************************************
//* ---------------------
//* S4CDCHitWire Class
//* ---------------------
//* (Description)
//*	Class for a single CDC hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4CDCHitWire		: Array containing "S4CDCHit"s
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include <iostream>
#include <iomanip>
#include <math.h>
#include "TRandom.h"
#include "S4CDCHitWire.h"
#include "S4CDCHit.h"
#include "S4CDCExactHit.h"

#define __TMP__
#define __TWOHITSEP__ 0.2
#define __TWOHITSEPZ__ 9999999999.

Double_t S4CDCHitWire::fgTwoHitSeparation  = __TWOHITSEP__;
Double_t S4CDCHitWire::fgTwoHitSeparationZ = __TWOHITSEPZ__;
Bool_t	 S4CDCHitWire::fgTPCMode = kFALSE;	

Double_t S4CDCHitWire::f2Tkpar[] = { -0.665417, 14.0052 };

ClassImp(S4CDCHitWire)

//=====================================================================
//* constructor -------------------------------------------------------

S4CDCHitWire::S4CDCHitWire(U4CDCTimeToXRelation *calib) 
	     : U4CDCHitWire(calib), fKilled(kFALSE)
{
}

//=====================================================================
//* destructor --------------------------------------------------------

S4CDCHitWire::~S4CDCHitWire()
{
	Delete();
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4CDCHitWire::DebugPrint(const Char_t *opt)
{
	std::cerr.setf(std::ios::right);
	std::cerr.setf(std::ios::fixed);
	std::cerr	<< "wire = (" << std::setw(1) << GetLayerNumber()
		<< "," << std::setw(1) << GetCellNumber()
		<< "," << std::setw(1) << GetWireNumber() << ")";
	TIter next(this);
	S4CDCHit *hp;
	Int_t n = 0;
	while ((hp = (S4CDCHit *)next())) {
		if (n != 0) std::cerr << "                  ";
		std::cerr << std::setw(3) << ++n << " ";
		hp->DebugPrint(opt);
	}
	std::cerr.unsetf(std::ios::right);
	std::cerr.unsetf(std::ios::fixed);
}

//=====================================================================
//* Append ------------------------------------------------------------

Int_t S4CDCHitWire::Append(S4CDCHit *ht)
{
	Int_t 	 layer, cell, wire;
	Double_t xwire, ywire;
	ChannelToWire(ht,layer,cell,wire,xwire,ywire);
	
	// Return value = -1:	1st hit.
	//		=  0:	successfully appended.
	//		=  1:	new wire.
	//		=  2:	new cell.
	//		=  3:	new layer.
	
        S4CDCExactHit *exhitp = ht->GetExactHitPtr();
	if (GetEntries() == 0) {
		fLayerNumber = layer;
		fCellNumber  = cell;
		fWireNumber  = wire;
                fWireTop     = exhitp->GetWireTop();
                fWireBtm     = exhitp->GetWireBtm();
                fTan2Stereo  = exhitp->GetTan2Stereo();
                fRwaist      = exhitp->GetRwaist();
		fXwire	     = xwire;
		fYwire	     = ywire;
      
		ht->fWirePtr = this;
		ht->Update();
                Add(ht);
                return -1;
	} else {
		if (layer == GetLayerNumber()) {
			if (cell == GetCellNumber()) {
                    		if (wire == GetWireNumber()) {
                    			ht->fWirePtr = this;
  					ht->Update();
                      			Add(ht);
                        		return 0;
                        	} else {
                        		return 1;
                        	}
                        } else {
                        	return 2;
                        }
                } else {
                        return 3;
                }
        }
}

//=====================================================================
//* ChannelToWire -----------------------------------------------------

void S4CDCHitWire::ChannelToWire(S4CDCHit *hit,
  			         Int_t &layer, Int_t &cell, Int_t &wire, 
  			         Double_t &x, Double_t &y)
{
	S4CDCExactHit* exacthitp = hit->GetExactHitPtr();
	layer = exacthitp->GetLayerNumber();
	cell  = exacthitp->GetCellNumber();
	wire  = exacthitp->GetWireNumber();
	x     = exacthitp->GetWireX();
	y     = exacthitp->GetWireY();

#if 0
	std::cerr << "S4CDCHitWire::ChannelToWire: layer = "
	     << layer << " cell = " << cell << " wire = "
	     << wire << " x = " << x << " y = " << y << std::endl;	
#endif
}

//=====================================================================
//* KillHits ----------------------------------------------------------

void S4CDCHitWire::KillHits( ) 
{
	UnSort();
	
	if (!fgTPCMode) {
		S4CDCHit::SetSortBy("TrueDriftLength");
		Sort();
		Double_t dllast    = -9999999.;
		Double_t zlast     = -9999999.;
		S4CDCHit *lasthitp = 0;
		Double_t distance  = dllast;
		Double_t zdistance = zlast;
	
		TIter next(this);
		S4CDCHit *hp;
		while ((hp = (S4CDCHit *)next())) {
			hp->SetLastHitPtr(lasthitp);
			Double_t dlcur = TMath::Abs(hp->GetExactHitPtr()->GetDriftLen());
			distance = dlcur - dllast;
			hp->SetDistanceToPrevious( distance );

			Double_t zcur  = (hp->GetExactHitPtr()->GetPosition()).Z();
			zdistance= TMath::Abs(zcur - zlast);		
#if 0
//			if ( distance < fgTwoHitSeparation && zdistance < fgTwoHitSeparationZ ) 
			if ( distance < fgTwoHitSeparation || zdistance < fgTwoHitSeparationZ ) 
			{
//				hp->SetBit(kECannotSeparate);
			}
#endif
//-------------------------------------------------------------------------------------------
//  TEMPORARY TREATMENT: Need to check if the last hit is really a delta ray.
//-------------------------------------------------------------------------------------------
			if (lasthitp && 
			   lasthitp->GetExactHitPtr()->GetPDGCode() == 11 &&
			   lasthitp->GetExactHitPtr()->GetMotherTrackID() == hp->GetTrackID()) {
				hp->SetBit(kEModifiedByDeltaRay);
				lasthitp->SetBit(kECannotSeparate);
	                        TVector3 pos = lasthitp->GetExactHitPtr()->GetPostPosition();
	                        distance = dlcur - pos.x();
	                        hp->SetDistanceToPrevious( distance );
		        }
//-------------------------------------------------------------------------------------------
			Double_t eff = (TMath::TanH((distance-f2Tkpar[0])*f2Tkpar[1])
					-TMath::TanH(-f2Tkpar[0]*f2Tkpar[1]))
                                         /(1.-TMath::TanH(-f2Tkpar[0]*f2Tkpar[1]));
			Double_t rand = gRandom->Uniform(0.,1.);
			if ( rand > eff )
			{
#if 0
                                std::cerr << " S4BabyHit::Kill:distance = " 
					  << distance << std::endl;
                                std::cerr << " S4BabyHit::Kill:Trackno =  " 
					  << hp->GetTrackID() << std::endl;
                                std::cerr << " S4BabyHit::Kill:Wireno =  " 
					  << hp->GetWireNumber() << std::endl;
                                std::cerr << " S4BabyHit::Kill:Cellno =  " 
					  << hp->GetCellNumber() << std::endl;
#endif
	                        hp->SetBit(kECannotSeparate);
                        }

//-------------------------------------------------------------------------------------------
//  TEMPORARY TREATMENT: Need to check if the last hit is really a delta ray.
//-------------------------------------------------------------------------------------------
			if (!lasthitp
			    || (lasthitp && lasthitp->GetExactHitPtr()->GetMotherTrackID() != hp->GetTrackID())) {

			        dllast = dlcur;
			        zlast  = zcur;
			}
//-------------------------------------------------------------------------------------------
			lasthitp = hp;
		}
	} else {
		S4CDCHit::SetSortBy("ZCoodinate");      // wrong sort ( no central membrane )
		Sort();
		Double_t dllast    = -9999999.;
		Double_t zlast     = -9999999.;
		Double_t distance  = dllast;
		Double_t zdistance = zlast;
	
		TIter next(this);
		S4CDCHit *hp;
		while ((hp = (S4CDCHit *)next())) {
			Double_t zcur = (hp->GetExactHitPtr()->GetPosition()).Z();
			zdistance= zcur - zlast;		
			hp->SetDistanceToPrevious( distance );

			Double_t dlcur  = hp->GetExactHitPtr()->GetDriftLen();
			distance = TMath::Abs(dlcur - dllast);
			if ( zdistance < fgTwoHitSeparationZ && distance < fgTwoHitSeparation ) 
			{
				hp->SetBit(kECannotSeparate);
			}
			dllast = dlcur;	
			zlast  = zcur;	
		}
	}	
	fKilled = kTRUE;
}	
