//*************************************************************************
//* ---------------------------
//* U4CDCTimeToXRelation Class
//* ---------------------------
//*************************************************************************

#include <iostream>
#include <iomanip>
#include <fstream>
#include "U4CDCTimeToXRelation.h"


Double_t U4CDCXTDatum::fSigmaMS = 0.;

// ---------------------
// X-t Calibration Datum
// ---------------------
ClassImp(U4CDCXTDatum)

U4CDCXTDatum::U4CDCXTDatum(Int_t n, Double_t *x, Double_t *t) : fN(n), fNpar(0)
{
        if (fN) {
                fX = new Double_t [fN];
                fT = new Double_t [fN];
                Int_t i;
                for (i = 0; i<fN; i++) {
                        fX[i] = x[i];
                        fT[i] = t[i];
                }
        } else {
                fX = 0;
                fT = 0;
        }
        fDXpar = 0; // DXpar has to be initialized by SetDXpar
}

U4CDCXTDatum::U4CDCXTDatum(U4CDCXTDatum &xt) : fN(xt.fN), fNpar(xt.fNpar)
{
        if (fN) {
                fX = new Double_t [fN];
                fT = new Double_t [fN];
                Int_t i;
                for (i = 0; i<fN; i++) {
                        fX[i] = xt.fX[i];
                        fT[i] = xt.fT[i];
                }
                fDXpar = new Double_t [fNpar];
                for (i = 0; i<4; i++) fDXpar[i] = xt.fDXpar[i];
        } else {
                fX = 0;
                fT = 0;
                fDXpar = 0;
        }
}

U4CDCXTDatum::~U4CDCXTDatum()
{
        if (fX)         delete [] fX;
        if (fT)         delete [] fT;
        if (fDXpar)     delete [] fDXpar;
}

void U4CDCXTDatum::SetDXpar(Int_t n, Double_t *par)
{
        if (n <= 0) {
                std::cerr    << "////// ERROR in U4CDCXTDatum::SetDXpar" << std::endl
                        << "  Invalid No points = " << n << std::endl;
                return;
        }
        if (n > fNpar) {
                if (fDXpar) delete [] fDXpar;
                fNpar  = n;
                fDXpar = new Double_t [fNpar];
        }
        Int_t i;
        for (i = 0; i<fNpar; i++) fDXpar[i] = par[i];
}

void U4CDCXTDatum::SetSigmaMS(Double_t sigma)
{
        fSigmaMS = sigma;
}


Int_t U4CDCXTDatum::WhereIs(Double_t t)
{
        Int_t base = 0;
        Int_t last = fN - 2;
        if (last < 0) {
                std::cerr << "U4CDCTimeToXRelation::CalibDatum::WhereIs()"
                     << " Negative last = " << last << std::endl;
                return 0;
        }
        Int_t position = 0;
        if (t <= fT[0] ) {
                position = 0;
        } else if (t >= fT[last]) {
                position = last;
        } else {
                while (last >= base) {
                        position = (base+last)/2;
                        if (fT[position] <= t && t < fT[position+1]) {
                                break;
                        } else if (fT[position+1] <= t) {
                                base = position + 1;
                        } else {
                                last = position;
                        }
                }
        }
        return position;
}

Double_t U4CDCXTDatum::X(Double_t t)
{
        Int_t n    = WhereIs(t);
        return fX[n] + ((fX[n+1]-fX[n])/(fT[n+1]-fT[n]))*(t-fT[n]);
}

Double_t U4CDCXTDatum::DX(Double_t x)
{
// START VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP
#if 0
        Double_t dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
                                   + fDXpar[2]*x + fDXpar[3] );
#else
        Double_t dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
                                   + fDXpar[2]*x + fDXpar[3]
                                   + fSigmaMS*fSigmaMS);
#endif
// END VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP
        return dx;
}

Double_t U4CDCXTDatum::Vd(Double_t t)
{
        Int_t n    = WhereIs(t);
        return (fX[n+1]-fX[n])/(fT[n+1]-fT[n]);
}

void U4CDCXTDatum::TimeToHitInfo(Double_t  t,   // time    [ns]
                                 Double_t &x,   // x       [cm]
                                 Double_t &dx,  // sigma_x [cm]
                                 Double_t &vd)  // Vdrift  [cm/ns]
{
        Int_t n    = WhereIs(t);

        x = fX[n] + ((fX[n+1]-fX[n])/(fT[n+1]-fT[n]))*(t-fT[n]);

// START VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP
#if 0
        dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
                                   + fDXpar[2]*x + fDXpar[3] );
#else
        dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
                                   + fDXpar[2]*x + fDXpar[3]
                                   + fSigmaMS*fSigmaMS);
#endif
// END VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP

        vd = (fX[n+1]-fX[n])/(fT[n+1]-fT[n]);
}



// --------------------
// X-t Calibration Data
// --------------------

ClassImp(U4CDCTimeToXRelation)

U4CDCTimeToXRelation::U4CDCTimeToXRelation(Int_t n, U4CDCXTDatum *xt, Int_t last)
{
        if (n) {
                CalibDatum *datum = new CalibDatum(last);
                datum->Add(xt);
                Add(datum);
                fCurr = datum;
        }
}

U4CDCTimeToXRelation::U4CDCTimeToXRelation(Char_t * calibfile)
{
        std::cout << " Going to open a calib file " << calibfile << std::endl;
        ifstream in;
        in.open(calibfile,std::ios::in);
        std::cout << " Opened " << calibfile << std::endl;
        Int_t last = 99999999;
        Int_t n;
        Double_t *x   = new Double_t [1000];
        Double_t *t   = new Double_t [1000];
        Double_t *par = new Double_t [100];
        Double_t mton = 1000.;
        while ((in >> last >> n)) {
                Int_t np, wireno;
                CalibDatum *calp = new CalibDatum(last);
                Int_t  iwir, i, idummy;
                Char_t cdummy[1000];
                Double_t ddummy;
                for (iwir = 0; iwir < n; iwir++) {
                        in >> wireno >> np;
                        for (i = 0; i < np; i++) {
                                in >> idummy >> x[i] >> t[i];
                                t[i] *= mton;
                        }
                        calp->Add(new U4CDCXTDatum(np, x, t));
                }

                for (iwir = 0; iwir < n; iwir++) {
                        in >> wireno >> np;
                        for (i = 0; i < np; i++) {
                                in >> idummy >> cdummy >> par[i]
                                   >> ddummy >> ddummy >> ddummy;
                        }
                        ((U4CDCXTDatum *)calp->At(iwir))->SetDXpar(np, par);
                }
                Add(calp);
        }
        delete [] x;
        delete [] t;
        delete [] par;
        in.close();
}

U4CDCTimeToXRelation::~U4CDCTimeToXRelation()
{
        TIter next(this);
        CalibDatum *calp;
        while ((calp = (CalibDatum *)next())) {
                calp->Delete();
        }
        Delete();
}

void U4CDCTimeToXRelation::Update()
{
        Int_t run = gJSF->GetRunNumber();
        TIter next(this);
        CalibDatum *datum;
        while ((datum = (CalibDatum *)next())) {
                fCurr = datum;
                if (run <= fCurr->GetLastRun()) break;
        }
}

#if 0
ClassImp(U4CDCTimeToXRelation)

U4CDCTimeToXRelation::U4CDCTimeToXRelation(Char_t * calibfile)
{
	std::cout << " Going to open calib file " << calibfile << std::endl;
	ifstream in;
	in.open(calibfile,std::ios::in);
	std::cout << " Opened " << calibfile << std::endl;

}
#endif
